//
// Basic Shape
//
// $Id: basic_shape.cpp 41 2011-01-08 16:35:39Z bambooflow $
// $Rev: 41 $
// $Author: bambooflow $
//
#include "basic_shape.h"

#include <cstdlib>
#include <cstring>
#include <cmath>

#ifndef M_PI
#define M_PI 3.1415926535897932384626433832795
#endif

Model* createPlane( float halfX, float halfY )
{
  float sizeX = 1.0f * halfX;
  float sizeY = 1.0f * halfY;

  glm::uint numVertices = 4;
  glm::uint numIndices = 6;

  Vertex vertices[] = { // (x,0,z)
    { glm::vec3(-sizeX, 0.0f, -sizeY ), glm::vec3( 0.0f, 1.0f, 0.0f ), glm::vec2( 0.0f, 0.0f ) },
    { glm::vec3( sizeX, 0.0f, -sizeY ), glm::vec3( 0.0f, 1.0f, 0.0f ), glm::vec2( 0.0f, 1.0f ) },
    { glm::vec3( sizeX, 0.0f,  sizeY ), glm::vec3( 0.0f, 1.0f, 0.0f ), glm::vec2( 1.0f, 1.0f ) },
    { glm::vec3(-sizeX, 0.0f,  sizeY ), glm::vec3( 0.0f, 1.0f, 0.0f ), glm::vec2( 1.0f, 0.0f ) }  
  };
  glm::uint indices[] = {
    0, 2, 1,
    0, 3, 2
  };

  return createShape(vertices, numVertices, indices, numIndices);
}


Model* createCube( float half )
{
  float sx = 1.0f * half;
  float sy = 1.0f * half;
  float sz = 1.0f * half;

  glm::uint numVertices = 24;
  glm::uint numIndices = 36;

  Vertex vertices[] = {
    { glm::vec3(-sx, -sy, -sz), glm::vec3(+0.0f, -1.0f, +0.0f), glm::vec2(0.0f, 0.0f) },
    { glm::vec3(-sx, -sy, +sz), glm::vec3(+0.0f, -1.0f, +0.0f), glm::vec2(0.0f, 1.0f) },
    { glm::vec3(+sx, -sy, +sz), glm::vec3(+0.0f, -1.0f, +0.0f), glm::vec2(1.0f, 1.0f) },
    { glm::vec3(+sx, -sy, -sz), glm::vec3(+0.0f, -1.0f, +0.0f), glm::vec2(1.0f, 0.0f) },
    { glm::vec3(-sx, +sy, -sz), glm::vec3(+0.0f, +1.0f, +0.0f), glm::vec2(1.0f, 0.0f) },
    { glm::vec3(-sx, +sy, +sz), glm::vec3(+0.0f, +1.0f, +0.0f), glm::vec2(1.0f, 1.0f) },
    { glm::vec3(+sx, +sy, +sz), glm::vec3(+0.0f, +1.0f, +0.0f), glm::vec2(0.0f, 1.0f) },
    { glm::vec3(+sx, +sy, -sz), glm::vec3(+0.0f, +1.0f, +0.0f), glm::vec2(0.0f, 0.0f) },
    { glm::vec3(-sx, -sy, -sz), glm::vec3(+0.0f, +0.0f, -1.0f), glm::vec2(0.0f, 0.0f) },
    { glm::vec3(-sx, +sy, -sz), glm::vec3(+0.0f, +0.0f, -1.0f), glm::vec2(0.0f, 1.0f) },
    { glm::vec3(+sx, +sy, -sz), glm::vec3(+0.0f, +0.0f, -1.0f), glm::vec2(1.0f, 1.0f) },
    { glm::vec3(+sx, -sy, -sz), glm::vec3(+0.0f, +0.0f, -1.0f), glm::vec2(1.0f, 0.0f) },
    { glm::vec3(-sx, -sy, +sz), glm::vec3(+0.0f, +0.0f, +1.0f), glm::vec2(0.0f, 0.0f) },
    { glm::vec3(-sx, +sy, +sz), glm::vec3(+0.0f, +0.0f, +1.0f), glm::vec2(0.0f, 1.0f) },
    { glm::vec3(+sx, +sy, +sz), glm::vec3(+0.0f, +0.0f, +1.0f), glm::vec2(1.0f, 1.0f) },
    { glm::vec3(+sx, -sy, +sz), glm::vec3(+0.0f, +0.0f, +1.0f), glm::vec2(1.0f, 0.0f) },
    { glm::vec3(-sx, -sy, -sz), glm::vec3(-1.0f, +0.0f, +0.0f), glm::vec2(0.0f, 0.0f) },
    { glm::vec3(-sx, -sy, +sz), glm::vec3(-1.0f, +0.0f, +0.0f), glm::vec2(0.0f, 1.0f) },
    { glm::vec3(-sx, +sy, +sz), glm::vec3(-1.0f, +0.0f, +0.0f), glm::vec2(1.0f, 1.0f) },
    { glm::vec3(-sx, +sy, -sz), glm::vec3(-1.0f, +0.0f, +0.0f), glm::vec2(1.0f, 0.0f) },
    { glm::vec3(+sx, -sy, -sz), glm::vec3(+1.0f, +0.0f, +0.0f), glm::vec2(0.0f, 0.0f) },
    { glm::vec3(+sx, -sy, +sz), glm::vec3(+1.0f, +0.0f, +0.0f), glm::vec2(0.0f, 1.0f) },
    { glm::vec3(+sx, +sy, +sz), glm::vec3(+1.0f, +0.0f, +0.0f), glm::vec2(1.0f, 1.0f) },
    { glm::vec3(+sx, +sy, -sz), glm::vec3(+1.0f, +0.0f, +0.0f), glm::vec2(1.0f, 0.0f) }
  };

  glm::uint indices[] = {
    0, 2, 1,
    0, 3, 2,
    4, 5, 6,
    4, 6, 7,
    8, 9, 10,
    8, 10, 11,
    12, 15, 14,
    12, 14, 13,
    16, 17, 18,
    16, 18, 19,
    20, 23, 22,
    20, 22, 21
  };

  return createShape(vertices, numVertices, indices, numIndices);
}


Model* createSphere( float radius, unsigned int numSlices )
{
  unsigned int numParallels = numSlices;
  unsigned int numVertices = ( numParallels + 1 ) * ( numSlices + 1 );
  unsigned int numIndices = numParallels * numSlices * 6;
  float angleStep = (2.0f * M_PI) / ((float) numSlices);

  Vertex* vertices = new Vertex[numVertices];
  if (!vertices) { return NULL; }

  glm::uint* indices = new glm::uint[numIndices];
  if (!indices) { return NULL; }

  for (unsigned i=0; i<numParallels+1; ++i) {
    for (unsigned j=0; j<numSlices+1; ++j) {
      glm::uint index = (i * (numSlices + 1) + j); 

      vertices[index].position.x = radius * sinf(0.5f * angleStep * (float)i) * sinf(angleStep * (float)j);
      vertices[index].position.y = radius * cosf(0.5f * angleStep * (float)i);
      vertices[index].position.z = radius * sinf(0.5f * angleStep * (float)i) * cosf(angleStep * (float)j);

      vertices[index].normal.x = vertices[index].position.x / radius;
      vertices[index].normal.y = vertices[index].position.y / radius;
      vertices[index].normal.z = vertices[index].position.z / radius;

      vertices[index].texcoord.s = (float)j / (float)numSlices;
      vertices[index].texcoord.t = (1.0f - (float)i) / (float)(numParallels - 1);
    }
  }

  glm::uint* indexBuf = indices;
  for (unsigned i=0; i<numParallels ; ++i) {
    for (unsigned j=0; j<numSlices; ++j) {
      *indexBuf++  = i * (numSlices + 1) + j;
      *indexBuf++ = (i + 1) * (numSlices + 1) + j;
      *indexBuf++ = (i + 1) * (numSlices + 1) + (j + 1);

      *indexBuf++ = i * (numSlices + 1) + j;
      *indexBuf++ = (i + 1) * (numSlices + 1) + (j + 1);
      *indexBuf++ = i * (numSlices + 1) + (j + 1);
    }
  }

  Model* shape = createShape(vertices, numVertices, indices, numIndices);

  delete [] vertices;
  delete [] indices;

  return shape;
}

Model* createTorus( float innerRadius, float outerRadius, unsigned int numSides, unsigned numFaces )
{
  glm::uint numVertices;
  glm::uint numIndices;

  if ((numSides<3) || (numFaces<3)) { return NULL; }

  float tIncr = 1.0f/(float)numFaces;
  float sIncr = 1.0f/(float)numSides;

  numVertices = (numFaces+1) * (numSides+1);
  numIndices  = numFaces * numSides * 2 * 3;

  Vertex* vertices = new Vertex[numVertices];
  if (!vertices) { return NULL; }

  glm::uint* indices = new glm::uint[numIndices];
  if (!indices) { return NULL; }

  float s = 0.0f;          
  for (unsigned i = 0; i <= numSides; i++, s+=sIncr) {
    float cos2PIs = (float)cosf(2.0f*M_PI*s);  
    float sin2PIs = (float)sinf(2.0f*M_PI*s);

    float t = 0.0f;
    for (unsigned j = 0; j <= numFaces; j++, t+=tIncr) {
      float cos2PIt = (float)cosf(2.0f*M_PI*t);
      float sin2PIt = (float)sinf(2.0f*M_PI*t);
      unsigned index = ((i * (numFaces +1)) + j);

      vertices[index].position.x = (outerRadius + innerRadius * cos2PIt) * cos2PIs;
      vertices[index].position.y = innerRadius * sin2PIt;
      vertices[index].position.z = (outerRadius + innerRadius * cos2PIt) * sin2PIs;

      vertices[index].normal.x = cos2PIs * cos2PIt;
      vertices[index].normal.y = sin2PIt;
      vertices[index].normal.z = sin2PIs * cos2PIt;

      vertices[index].texcoord.s = s;
      vertices[index].texcoord.t = t;
    }
  }

  unsigned indexIndices = 0;
  for (unsigned i = 0; i < numSides; ++i) {
    for (unsigned j = 0; j < numFaces; ++j) {
      float v0 = ((i * (numFaces +1)) + j);    
      float v1 = ((i * (numFaces +1)) + (j+1));
      float v2 = (((i+1) * (numFaces +1)) + (j+1));
      float v3 = (((i+1) * (numFaces +1)) + j);

      indices[indexIndices++] = v0;
      indices[indexIndices++] = v1;
      indices[indexIndices++] = v2;

      indices[indexIndices++] = v0;
      indices[indexIndices++] = v2;
      indices[indexIndices++] = v3;
    }
  }
  Model* shape = createShape(vertices, numVertices, indices, numIndices);

  delete [] vertices;
  delete [] indices;

  return shape;
}

Model* createCapsule( float radius, float length, unsigned int numSlices )
{
  glm::uint numParallels = numSlices;
  glm::uint numVertices = ( numParallels + 2 ) * ( numSlices + 1 );
  glm::uint numIndices = (numParallels+1) * numSlices * 6;
  float angleStep = (2.0f * M_PI) / ((float) numSlices);
  float l = length * 0.5;

  Vertex* vertices = new Vertex[numVertices];
  if (!vertices) { return NULL; }

  glm::uint* indices = new glm::uint[numIndices];
  if (!indices) { return NULL; }

  unsigned index = 0;
  for (unsigned i=0; i<numParallels/2; ++i) {
    // cylinder top cap
    for (unsigned j=0; j<numSlices+1; ++j) {
      vertices[index].position.x = radius * sinf(0.5f*angleStep * (float)i) * sinf(angleStep * (float)j);
      vertices[index].position.y = radius * cosf(0.5f*angleStep * (float)i) + l;
      vertices[index].position.z = radius * sinf(0.5f*angleStep * (float)i) * cosf(angleStep * (float)j);

      vertices[index].normal.x = sinf(0.5f*angleStep * (float)i) * sinf(angleStep * (float)j);
      vertices[index].normal.y = cosf(0.5f*angleStep * (float)i);
      vertices[index].normal.z = sinf(0.5f*angleStep * (float)i) * cosf(angleStep * (float)j);

      vertices[index].texcoord.s = sinf(0.5f*angleStep * (float)i) * sinf(angleStep * (float)j);
      vertices[index].texcoord.t = sinf(0.5f*angleStep * (float)i) * cosf(angleStep * (float)j);
      ++index;
    }
  }
  {
    // cylinder body
    for (unsigned j=0; j<numSlices+1; ++j) {
      vertices[index].position.x = radius * sinf(angleStep*(float)j);
      vertices[index].position.y = l;
      vertices[index].position.z = radius * cosf(angleStep*(float)j);

      vertices[index].normal.x = sinf(angleStep*(float)j);
      vertices[index].normal.y = 0.0f;
      vertices[index].normal.z = cosf(angleStep*(float)j);

      vertices[index].texcoord.s = sinf(angleStep*(float)j);
      vertices[index].texcoord.t = cosf(angleStep*(float)j);
      ++index;
    }
    for (unsigned j=0; j<numSlices+1; ++j) {
      vertices[index].position.x = radius * sinf(angleStep*(float)j);
      vertices[index].position.y = -l;
      vertices[index].position.z = radius * cosf(angleStep*(float)j);

      vertices[index].normal.x = sinf(angleStep*(float)j);
      vertices[index].normal.y = 0.0f;
      vertices[index].normal.z = cosf(angleStep*(float)j);

      vertices[index].texcoord.s = sinf(angleStep*(float)j);
      vertices[index].texcoord.t = cosf(angleStep*(float)j);

      ++index;
    }
  }
  for (unsigned i=numParallels/2+1; i<numParallels+1; ++i) {
    // cylinder bottom cap
    for (unsigned j=0; j<numSlices+1; ++j) {
      vertices[index].position.x = radius * sinf (0.5f*angleStep*(float)i) * sinf(angleStep*(float)j);
      vertices[index].position.y = radius * cosf (0.5f*angleStep*(float)i) - l;
      vertices[index].position.z = radius * sinf (0.5f*angleStep*(float)i) * cosf(angleStep*(float)j);

      vertices[index].normal.x = sinf (0.5f*angleStep*(float)i) * sinf(angleStep*(float)j);
      vertices[index].normal.y = cosf (0.5f*angleStep*(float)i);
      vertices[index].normal.z = sinf (0.5f*angleStep*(float)i) * cosf(angleStep*(float)j);

      vertices[index].texcoord.x = sinf (0.5f*angleStep*(float)i) * sinf(angleStep*(float)j);
      vertices[index].texcoord.y = sinf (0.5f*angleStep*(float)i) * cosf(angleStep*(float)j);

      ++index;
    }
  }

  unsigned int* indexBuf = indices;
  for (unsigned i=0; i<numParallels+1; ++i) {
    for (unsigned j=0; j<numSlices; ++j) {
      *indexBuf++ = i       * (numSlices + 1) + j;
      *indexBuf++ = (i + 1) * (numSlices + 1) + j;
      *indexBuf++ = (i + 1) * (numSlices + 1) + (j + 1);

      *indexBuf++ = i       * (numSlices + 1) + j;
      *indexBuf++ = (i + 1) * (numSlices + 1) + (j + 1);
      *indexBuf++ = i       * (numSlices + 1) + (j + 1);
    }
  }

  Model* shape = createShape(vertices, numVertices, indices, numIndices);

  delete [] vertices;
  delete [] indices;

  return shape;
}


Model* createCylinder( float radius, float length, unsigned int numSlices )
{
  unsigned int numVertices = 1 + (numSlices+1)*4 + 1;
  unsigned int numIndices = (numSlices + numSlices*4 + numSlices) * 3;
  float angleStep = (2.0f * M_PI) / ((float) numSlices);
  float l = length * 0.5;

  Vertex* vertices = new Vertex[numVertices];
  if (!vertices) { return NULL; }

  glm::uint* indices = new glm::uint[numIndices];
  if (!indices) { return NULL; }


  unsigned int index = 0;
  {
    // cylinder top cap
    {
      // center
      vertices[index].position.x = 0.0f;
      vertices[index].position.y = l;
      vertices[index].position.z = 0.0f;

      vertices[index].normal.x = 0.0f;
      vertices[index].normal.y = 1.0f;
      vertices[index].normal.z = 0.0f;

      vertices[index].texcoord.s = 0.0f;
      vertices[index].texcoord.t = 0.0f;

      ++index;
    }
    for (unsigned j=0; j<numSlices+1; ++j) {
      vertices[index].position.x = radius * sinf(angleStep * (float)j);
      vertices[index].position.y = l;
      vertices[index].position.z = radius * cosf(angleStep * (float)j);

      vertices[index].normal.x = 0.0f;
      vertices[index].normal.y = 1.0f;
      vertices[index].normal.z = 0.0f;

      vertices[index].texcoord.s = sinf(angleStep * (float)j);
      vertices[index].texcoord.t = cosf(angleStep * (float)j);

      ++index;
    }
  }
  {
    // cylinder body (top)
    for (unsigned j=0; j<numSlices+1; ++j) {
      vertices[index].position.x = radius * sinf(angleStep * (float)j);
      vertices[index].position.y = l;
      vertices[index].position.z = radius * cosf(angleStep * (float)j);

      vertices[index].normal.x = sinf(angleStep * (float)j);
      vertices[index].normal.y = 0.0f;
      vertices[index].normal.z = cosf(angleStep * (float)j);

      vertices[index].texcoord.s = sinf(angleStep * (float)j);
      vertices[index].texcoord.t = cosf(angleStep * (float)j);

      ++index;
    }
    // cylinder body (bottom)
    for (unsigned j=0; j<numSlices+1; ++j) {
      vertices[index].position.x = radius * sinf(angleStep * (float)j);
      vertices[index].position.y = -l;
      vertices[index].position.z = radius * cosf(angleStep * (float)j);

      vertices[index].normal.x = sinf(angleStep * (float)j);
      vertices[index].normal.y = 0.0f;
      vertices[index].normal.z = cosf(angleStep * (float)j);

      vertices[index].texcoord.s = sinf(angleStep * (float)j);
      vertices[index].texcoord.t = cosf(angleStep * (float)j);

      ++index;
    }
  }
  {
    // cylinder bottom cap

    for (unsigned j=0; j<numSlices+1; ++j) {
      vertices[index].position.x = radius * sinf (angleStep * (float)j);
      vertices[index].position.y = -l;
      vertices[index].position.z = radius * cosf (angleStep * (float)j);

      vertices[index].normal.x =  0.0f;
      vertices[index].normal.y = -1.0f;
      vertices[index].normal.z =  0.0f;

      vertices[index].texcoord.s = sinf (angleStep * (float)j);
      vertices[index].texcoord.t = cosf (angleStep * (float)j);

      ++index;
    }
    {
      // center
      vertices[index].position.x = 0.0f;
      vertices[index].position.y = -l;
      vertices[index].position.z = 0.0f;

      vertices[index].normal.x =  0.0f;
      vertices[index].normal.y = -1.0f;
      vertices[index].normal.z =  0.0f;

      vertices[index].texcoord.s = 0.0f;
      vertices[index].texcoord.t = 0.0f;

      ++index;
    }
  }

  unsigned int* indexBuf = indices;
  {
    // cylinder top cap
    unsigned center = 0;
    unsigned offset = 1;
    for (unsigned j=0; j<numSlices; ++j) {
      *indexBuf++ = center;
      *indexBuf++ = offset + j;
      *indexBuf++ = offset + (j + 1);
    }
  }
  {
    // cylinder body
    unsigned offset = 1 + (numSlices+1);
    for (unsigned j=0; j<numSlices; ++j) {
      *indexBuf++ = offset + j;
      *indexBuf++ = offset + (numSlices + 1) + j;
      *indexBuf++ = offset + (numSlices + 1) + (j + 1);

      *indexBuf++ = offset + j;
      *indexBuf++ = offset + (numSlices + 1) + (j + 1);
      *indexBuf++ = offset + (j + 1);
    }
  }
  {
    // cylinder bottom cap
    unsigned center = 1 + (numSlices+1)*4;
    unsigned offset = 1 + (numSlices+1)*3;
    // cylinder top cap
    for (unsigned j=0; j<numSlices; ++j) {
      *indexBuf++ = center;
      *indexBuf++ = offset + (j + 1);
      *indexBuf++ = offset + j;
    }
  }

  Model* shape = createShape(vertices, numVertices, indices, numIndices);

  delete [] vertices;
  delete [] indices;

  return shape;
}



Model* createCone( float radius, float length, unsigned int numSlices )
{
  unsigned int numVertices = 1 + (numSlices+1)*2 + 1;
  unsigned int numIndices = (numSlices + numSlices*2 + numSlices) * 3;
  float angleStep = (2.0f * M_PI) / ((float) numSlices);
  float l = length * 0.5;

  Vertex* vertices = new Vertex[numVertices];
  if (!vertices) { return NULL; }

  glm::uint* indices = new glm::uint[numIndices];
  if (!indices) { return NULL; }

  unsigned int index = 0;
  {
    // cylinder top cap
    {
      // center
      vertices[index].position.x = 0.0f;
      vertices[index].position.y = l;
      vertices[index].position.z = 0.0f;

      vertices[index].normal.x = 0.0f;
      vertices[index].normal.y = 1.0f;
      vertices[index].normal.z = 0.0f;

      vertices[index].texcoord.s = 0.0f;
      vertices[index].texcoord.t = 0.0f;

      ++index;
    }
    // cylinder body (bottom)
    for (unsigned j=0; j<numSlices+1; ++j) {
      vertices[index].position.x = radius * sinf (angleStep * (float)j);
      vertices[index].position.y = -l;
      vertices[index].position.z = radius * cosf (angleStep * (float)j);

      vertices[index].normal.x = sinf (angleStep * (float)j);
      vertices[index].normal.y = 0.0f;
      vertices[index].normal.z = cosf (angleStep * (float)j);

      vertices[index].texcoord.s = sinf (angleStep * (float)j);
      vertices[index].texcoord.t = cosf (angleStep * (float)j);

      ++index;
    }
  }
  {
    // cylinder bottom cap
    for (unsigned j=0; j<numSlices+1; ++j) {
      vertices[index].position.x = radius * sinf (angleStep * (float)j);
      vertices[index].position.y = -l;
      vertices[index].position.z = radius * cosf (angleStep * (float)j);

      vertices[index].normal.x =  0.0f;
      vertices[index].normal.y = -1.0f;
      vertices[index].normal.z =  0.0f;

      vertices[index].texcoord.s = sinf (angleStep * (float)j);
      vertices[index].texcoord.t = cosf (angleStep * (float)j);

      ++index;
    }
    {
      // center
      vertices[index].position.x = 0.0f;
      vertices[index].position.y = -l;
      vertices[index].position.z = 0.0f;

      vertices[index].normal.x =  0.0f;
      vertices[index].normal.y = -1.0f;
      vertices[index].normal.z =  0.0f;

      vertices[index].texcoord.s = 0.0f;
      vertices[index].texcoord.t = 0.0f;

      ++index;
    }
  }

  unsigned int* indexBuf = indices;
  {
    // cylinder top cap
    unsigned center = 0;
    unsigned offset = 1;
    for (unsigned j=0; j<numSlices; ++j) {
      *indexBuf++ = center;
      *indexBuf++ = offset + j;
      *indexBuf++ = offset + (j + 1);
    }
  }
  {
    // cylinder bottom cap
    unsigned center = 1 + (numSlices+1)*2;
    unsigned offset = 1 + (numSlices+1)*1;
    // cylinder top cap
    for (unsigned j=0; j<numSlices; ++j) {
      *indexBuf++ = center;
      *indexBuf++ = offset + (j + 1);
      *indexBuf++ = offset + j;
    }
  }

  Model* shape = createShape(vertices, numVertices, indices, numIndices);

  delete [] vertices;
  delete [] indices;

  return shape;
}


